; This is the code for searching the monster list for things that can be 'seen'
; by the men, and consequently which can also 'see' the men.

; This is done by trying to draw a line in the maze from the man to any monster
; within 'range' and checking every cell that the line includes for wall or door
; characters.

; This file ( S.Brattel )

; Given ds:si as a pointer to man 0 or man 1, find which monsters can see him
; and build a table of parameters defining the lines of view.

; NOTE, we assume that not more than 1/8 of the things can be seen at any one
; point, this is reasonable given the value of Look_Range but could be violated
; if the maze style is changed drastically.

; See the definition of Sight_WS_0/1 in main.

; Given ds:si = ptr to man

Look:		push es			; Save
		push di			;
		push si			;

		; Get data from the current man

		mov bl,mType[si]	; Get the man's number

		mov bh,mFlags[si]	; Get the man's flags

		mov dx,mOff[si]		; "X"
		mov ax,mSeg[si]		; "Y"

		mov es,ax		; Seg
		mov di,dx		; Off

		; Now get a pointer to the sight workspace

		mov bp,o Sight_WS_0	; Assume man 0
		mov si,mSightpMan0	;

		cmp bl,0		; Is it 0 ?
		je L1			; Yes, skip

		mov bp,o Sight_WS_1	; Man 1
		mov si,mSightpMan1	;

		; cs:bp is the workspace

L1:		test bh,080h		; Is the man dead ?
		jnz L_Dead		; Yes, skip

		; Point at the first thing

		mov cs:Thing_pMan,si	; Save ptr

		mov si,lMon		; After the two men (one for now)

		; Now calculate the clipping range for the offset and segment

		sub dx,Look_Range	; LHS of clipping box

		sub ax,Look_Range * BM_Y_Stride	; Top of clipping box

		; Get the number of things

		mov cx,nMons		; This many

		; Loop through the things, looking for each live one

L_Lp:		test b mFlags[si],80h	; Alive ?
		jne L_L_Next		; No, skip

		; Is its X in range ?

		mov bx,mOff[si]		; Giz its offset
		sub bx,dx		; Relative to clip-box lhs

		cmp bx,Look_Range * 2	; In box ?
		jb L_L_X		; Possibly, check further

		; No, skip (usually we get here, so faster this way)

		; Invisible, clear the ptr

L_L_Next_1:	mov bx,cs:Thing_pMan	; Get offset dep. on current man
		mov w [bx+si],0		; Clear ptr to line constants

		; Skip this monster

L_L_Next:	add si,lMon		; Skip monster

		loop L_Lp		; Loop

		; Now we check if other man is visible . . .

		; One day, at least . . .

		; Terminate the list

L_Exit:		mov b cs:sRange[bp],127	; Flag none

		; Done

		pop si			; Restore
		pop di			;
		pop es			;

		ret			; Done

		; If the man is dead, nobody can see him

L_Dead:		mov si,lMon		; Point at the first thing

		mov cx,nMons		; This many

		; Clear the remaining thing's ptrs

L_Term:		mov ax,0		; Can't see him

		; Clear the pointer to the line constants in the things WS.

		mov bx,cs:Thing_pMan	; Get offset dep. on current man

LD_Lp:		mov [bx+si],ax		; Save ptr to line constants

		add si,lMon		; Skip

		loop LD_Lp		; Loop

		; Done

		jmp s L_Exit		; Done

		; Is its Y in range ?

L_L_X:		mov bx,mSeg[si]		; Giz its seg.
		sub bx,ax		; Relative to clip-box top

		cmp bx,Look_Range * BM_Y_Stride * 2 ; In box ?
		jae L_L_Next_1			    ; No, fail

		; It is in the clip-box, now try line

		push dx			; Save dx
		push cx			;      cx
		push ax			;      ax
		push si			;      si

		; es:di is the man's position

		mov dx,mSeg[si]		; Seg of target
		mov cx,mOff[si]		; Off

		; Get ptr to the line constant store

		mov si,bp		; Point at some storage

		; Can we see it ?

		call Look_Line		; Try to see it

		; Restore bp from si

		mov bp,si		; Restore

		; Recover ptr to thing

		pop si			; Restore ptr

		; Did we see it ?

		jc L_L_Inv		; No, skip

		; We can see it, setup ptr to the thing

		mov cs:spTarget[bp],si	; Store ptr to target

		; Store the pointer to the line constants in the things WS.

		mov bx,cs:Thing_pMan	; Get offset dep. on current man
		mov [bx+si],bp		; Save ptr to line constants

		; Move past this entry

		add bp,lSight		; Skip line data

		; Recover the registers

		pop ax			; Recover
		pop cx			;
		pop dx			;

		; Now, is this the last entry ?

		; Check for end of either, since we don't know which we are in.

		cmp bp,o Sight_WS_E0	; End of first ?
		je L_Term		; Yes, end of search

		cmp bp,o Sight_WS_E1	; End of second ?
		je L_Term		; Yes, end of search

		; Try next thing

		jmp s L_L_Next		; Loop

		; This is not visible, clear the pLine constants

L_L_Inv:	mov bx,cs:Thing_pMan	; Get offset dep. on current man
		mov w [bx+si],0		; Clear ptr to line constants

		; Restore

		pop ax			; Recover
		pop cx			;
		pop dx			;

		; Try the next one

		jmp s L_L_Next		; Loop

; Given es:di as one position, dx:cx as another and cs:si as pointer to line
; workspace, calculate the line-drawing constants and see if the line is clear.

; Fail, no regs saved

LL_Fail:	stc			; Assert fail
		ret			;

Look_Line:	sub cx,di		; Calc cx := dX

		mov bp,1		; Assume step right (+ve)
		jns LL1			; +ve, skip

		neg cx			; Make dX positive,
		neg bp			; Give the X step the correct sign

LL1:		cmp cx,Look_Range	; Is it visible ?
		jae LL_Fail		; No, skip

		; Now same for Y

		mov ax,es		; Get Ys
		mov bx,dx		;

		sub bx,ax		; Calc bx := dY

		mov dx,BM_Y_Stride	; Assume step down (+ve)
		jns LL2			; +ve, skip

		neg bx			; Make dY positve
		neg dx			; Give the Y step the correct sign

LL2:		cmp bx,Look_Range * BM_Y_Stride	; Is it visible ?
		jae LL_Fail		; No, skip

		; We are going to have to try, save regs

		push es			; Save posn
		push di			;

		; Scale dy down

		push dx			; Save

		xor dx,dx		; 32 - bit

		mov ax,bx		; Get dY
		mov bx,BM_Y_Stride	;
		div bx			; Scale

		mov bx,ax		; Return scaled value

		pop dx			; Restore

		; cx = dX, bp = X step ; bx = dY, dx = Y step

		cmp cx,bx		; dX > dY ?
		jae LL3			; Yes, skip

		; dY > dX so we store dY as horiz

		mov w cs:sdHoriz[si],0	; dOffset := 0
		mov cs:sdHoriz+2[si],dx	; dSeg    := Y step

		; Swap dX and dY

		xchg cx,bx		; Swap dX and dY

		jmp s LL4		; Skip

		; dX > dY so we store dX as horiz

LL3:		mov cs:sdHoriz[si],bp	; dOffset := X step
		mov w cs:sdHoriz+2[si],0; dSeg	  := 0

		; Store diag step

LL4:		mov cs:sdDiag[si],bp	; X step
		mov cs:sdDiag+2[si],dx	; Y step

		; Now calculate the error term / constants

		mov bh,bl		; Get dY
		add bh,bh		; bh := 2dY (c2)

		mov ah,bh		; Get 2dY
		sub ah,cl		; 2dY-dx (e)

		mov bl,ah		; 2dY-dx
		sub bl,cl		; 2dY-2dX (c1)

		; Store constants

		mov cs:s_e[si],ah	; E
		mov cs:s_c1[si],bl	; C1
		mov cs:s_c2[si],bh	; C2
		mov cs:sRange[si],cl	; dX (Range)

		; Perform the line algorithm.

		; First, are we moving up or down ?

		test dh,80h		; Test sign
		jz LL5			; +ve, we are moving down

		; Moving up, es:di will not do, we must make di as large as
		; possible so that we can move up by decrementing the offset.

		add di,0F000h		; Increase offset
		mov dx,es		; Decrease segment
		sub dx,0F00h		;
		mov es,dx		;

		; es:di still points at the start, but in such a way that we
		; can step in Y by altering the offset

		; Calc Horiz vector

LL5:		mov dx,cs:sdHoriz+2[si]	; Giz dSeg
		add dx,dx		; * 16
		add dx,dx		;
		add dx,dx		;
		add dx,dx		;
		add dx,cs:sdHoriz[si]	; dx := Horiz displacement

		; Calc diag vector

		mov bp,cs:sdDiag+2[si]	; Giz dSeg
		add bp,bp		; * 16
		add bp,bp		;
		add bp,bp		;
		add bp,bp		;
		add bp,cs:sdDiag[si]	; bp := Diag displacement

		; Get 16-bit length

		mov ch,0		; cx := length

		; Draw the line

LL_Lp:		mov al,es:[di]		; Clear ?

		and al,0F0h		; Wall ?
		cmp al,090h		;
		jz LL_Fail_R		; Yes, fail

		; Next

		cmp ah,0		; e +ve ?
		jns LL_L1		; Yes, skip

		add ah,bh		; e := e + 2dY

		add di,dx		; Horiz

		loop LL_Lp		; Loop

		pop di			; Restore
		pop es			;

		clc			; Ok
		ret			;

LL_L1:		add ah,bl		; e := e + 2dY - 2dX

		add di,bp		; Diag

		loop LL_Lp		; Loop

		pop di			; Restore
		pop es			;

		clc			; Ok
		ret			;

LL_Fail_R:	pop di			; Restore
		pop es			;

		stc			; Ok
		ret			;

; Vars

Thing_pMan	dw ?			; Temp offset store